pro overlay, a, xx, yy, NLEVELS = nlev, levels=levels, TITLES = titles, $
                        NOASPECT= noaspect, WINDOW_SCALE = window_scale, $
			GRID = grid, CONTOURSMOOTH = contoursmooth, $
			RANGE = range, radec=radec, xlog=xlog, ylog=ylog 
 
;+
; NAME:
;	OVERLAY
;
; PURPOSE:
;	Overlay an image and a contour plot.
;
; CATEGORY:
;	General graphics.
;
; CALLING SEQUENCE:
;	OVERLAY, A [, XX, YY]
;
; INPUTS:
;	A:	The two-dimensional array to display.
;	XX,YY:	The fields containing the axis grids.
;
; KEYWORD PARAMETERS:
; WINDOW_SCALE:	Set this keyword to scale the window size to the image size.
;		Otherwise, the image size is scaled to the window size.
;		This keyword is ignored when outputting to devices with 
;		scalable pixels (e.g., PostScript).
;
; OUTPUTS:
;	No explicit outputs.
;
; COMMON BLOCKS:
;	None.
;
; SIDE EFFECTS:
;	The currently selected display is affected.
;
; RESTRICTIONS:
;	None.
;
; PROCEDURE:
;	If the device has scalable pixels, then the image is written over
;	the plot window.
;
; MODIFICATION HISTORY:
;	DMS, May, 1988.
;-

;on_error,2                      ;Return to caller if an error occurs
sz = size(a)			;Size of image
if sz[0] lt 2 then message, 'Parameter not 2D'

	;set window used by contour
contour,[[0,0],[1,1]],/nodata, xstyle=4, ystyle = 4

; Obtain window settings
px = !x.window * !d.x_vsize	;Get size of window in device units
py = !y.window * !d.y_vsize
swx = (px[1]-px[0])*0.875	;Size in x in device units, add space for color bar
swy = py[1]-py[0]		;Size in Y
six = sz[1]			;Image sizes
siy = sz[2]
aspi = float(six) / siy		;Image aspect ratio
aspw = swx / swy		;Window aspect ratio
f = aspi / aspw			;Ratio of aspect ratios

; exception handling for NaNs in the data
valid=where(FINITE(a), count)
if count lt 1 then message, "No valid data points found."
validpart=a[valid]
maxval=max(validpart)
minval=min(validpart)

if n_elements(range) eq 2 then nanvalue=range[0] else begin
 epsilon=1.0/128.0
 nanvalue=minval-epsilon*(maxval-minval)
 if count lt six*siy then range=[nanvalue,maxval] else range=[minval,maxval]
endelse

b=make_array(size=size(a),value=nanvalue)
b[valid]=validpart

; Produce color plot
if (!d.flags and 1) ne 0 then begin	;Scalable pixels?
				;Adjust window size
  if not keyword_set(noaspect) then begin
    if f ge 1.0 then swy = swy / f else swx = swx * f
  endif

  tv,bytscl(b,min=range[0],max=range[1]),px[0],py[0],$
			xsize = swx, ysize = swy, /device

endif else begin	;Not scalable pixels	
   if keyword_set(window_scale) then begin ;Scale window to image?
	tv,bytscl(b,min=range[0],max=range[1]),px[0],py[0]	;Output image
	swx = float(six)	;Set window size from image
	swy = float(siy)
    endif else begin		;Scale window
        if not keyword_set(noaspect) then begin
          if f ge 1.0 then swy = swy / f else swx = swx * f
        endif
	tv,congrid(bytscl(b,min=range[0],max=range[1]),swx,swy),px[0],py[0]
	endelse			;window_scale
     endelse			;scalable pixels

mx = !d.n_colors-1		;Brightest color
				;color vectors

; Contour level definition
; default
nlevbas=n_elements(levels)
if n_elements(nlev) EQ 0 then begin
	nlev=nlevbas
	if nlev eq 0 then nlev=5
endif

; No contours for constant surface
if (maxval-minval) le 1e-5*abs(maxval+minval) then nlev=0

if nlev NE 0 then begin
  colors=intarr(nlev)
  for i=0,(nlev-1)/2 do colors[i]=mx
  for i=(nlev-1)/2+1,nlev-1 do colors[i]=0
  nodata=0
; Create really equidistant levels
  if nlevbas eq 0 then begin
    a0=float(range[0])
    da=(range[1]-a0)/(nlev+1)
    levels=fltarr(nlev)
    for j=0,nlev-1 do levels[j]=a0+(j+1)*da
  endif else begin
    levels=levels[0:nlev-1]
  endelse
; No contours
endif else begin
  colors=[mx]
  nodata=1
  levels=[float(range[0])]
endelse

; Create the color bar
steps = 100
xsize = swx/10
ysize = swy/steps
xpos = px[0] + swx +1
ypos = py[0]

a0=float(range[0])
da=(range[1]-a0)/(steps-1)
feld2=fltarr(2,steps)
for j=0,steps-1 do begin feld2[0,j]=a0+j*da & feld2[1,j]=feld2[0,j] & endfor

if (!d.flags and 1) ne 0 then begin	;Scalable pixels
  tv,bytscl(feld2,min=range[0],max=range[1]),xpos,ypos, $
		xsize = xsize, ysize = swy, /device
endif else begin			;Not scalable pixels	
for j=0,steps-1 do begin
  color=fix((float(!d.table_size)/steps)*j)
  tv, replicate(color,xsize,fix(ysize)+1), fix(xpos), fix(ypos+j*ysize)
endfor
endelse

; contours/lines in wedge
ax1=[0.0,1.0]
ax2=fltarr(steps)
for j=0,steps-1 do ax2[j]=a0+j*da
contour, feld2, ax1,ax2, /noerase, $
         pos=[xpos,ypos,fix(xpos+xsize),ypos+swy], /dev, $
         levels = levels, c_color = colors, nodata = nodata, ystyle=5, $
         xticks=1, xtickname=replicate(' ',2), xtickformat='(A1)', xminor=1 

if n_elements(titles) EQ 0 then titles=[' ',' ',' ']

; wedge axis
axis, yaxis=1, ytitle=titles[2], ystyle=1, ticklen=0.2, yminor=5

; Compute axes
nx=n_elements(xx)-1
ny=n_elements(yy)-1
; create axes if not given
if (nx lt 1) then begin
  xx=findgen(six)
  nx=six-1
endif
if (ny lt 1) then begin
  yy=findgen(siy)
  ny=siy-1
endif

; Logarithmic scale
if keyword_set(xlog) then xxcontour=alog10(xx) else xxcontour=xx
if keyword_set(ylog) then yycontour=alog10(yy) else yycontour=yy

; Limits
xx1=xxcontour[0]-(xxcontour[1]-xxcontour[0])*0.5
xx2=xxcontour[nx]+(xxcontour[nx]-xxcontour[nx-1])*0.5
yy1=yycontour[0]-(yycontour[1]-yycontour[0])*0.5
yy2=yycontour[ny]+(yycontour[ny]-yycontour[ny-1])*0.5

; smoothing step if desired
if n_elements(contoursmooth) gt 0 then begin
  deltax=(xxcontour[nx-1]-xxcontour[0])/(nx-1)
  deltay=(yycontour[ny-1]-yycontour[0])/(ny-1)
  if (n_elements(contoursmooth) eq 1) then begin
    if contoursmooth eq 1 then begin
      binsizex=3
      binsizey=3
    endif else begin
      binsizex=contoursmooth
      binsizey=contoursmooth
    endelse
  endif else begin
      binsizex=contoursmooth[0]
      binsizey=contoursmooth[1]
  endelse
  acontour=min_curve_surf(a,xgrid=[xxcontour[0],deltax],ygrid=[yycontour[0],deltay],$
					    nx=nx*binsizex,ny=ny*binsizey)
  xxcontour=rebin(xxcontour,nx*binsizex)
  yycontour=rebin(yycontour,ny*binsizey)
endif else acontour=a

; Logarithmic scale
linframe=[xx1,yy1,xx2,yy2]
if keyword_set(xlog) then begin
  xxcontour=10.^(xxcontour)
  xx1=10.^xx1
  xx2=10.^xx2
endif
if keyword_set(ylog) then begin
  yycontour=10.^(yycontour)
  yy1=10.^yy1
  yy2=10.^yy2
endif

; Now add the contour plot
if keyword_set(radec) then begin
  ticks=radeclabels(xx,yy)
  contour,acontour,xxcontour,yycontour,/noerase,/xst,/yst,$
	xrange=[xx1,xx2], yrange=[yy1,yy2],$
	pos = [px[0],py[0], px[0]+swx,py[0]+swy],/dev,$
	levels = levels, c_color =  colors, nodata = nodata, $
	xtitle=titles[0], ytitle=titles[1], $
        xticks=ticks.xTicks, yticks=ticks.yTicks, xtickv=ticks.xPos, $
        ytickv=ticks.yPos, xtickname=ticks.xNames, ytickname=ticks.yNames
endif else contour,acontour,xxcontour,yycontour,/noerase,/xst,/yst,$
	xrange=[xx1,xx2], yrange=[yy1,yy2],$
	pos = [px[0],py[0], px[0]+swx,py[0]+swy],/dev,$
	levels = levels, c_color =  colors, nodata = nodata, $
	xtitle=titles[0], ytitle=titles[1], xlog=xlog, ylog=ylog

; Title
xchar=!d.x_ch_size*(1.0+!p.charsize)
ychar=!d.y_ch_size*(1.0+!p.charsize)
if (n_elements(titles) gt 3) then xyouts, px[0]+2.0*xchar, py[0]+swy-1.0*ychar, /dev, titles[3], $
	   charsize=!p.charsize*1.2, color=30974, charthick=2.0

; Possible return grid
ax=px[0]
ay=py[0]
wedgemax=a0+(steps-1)*da
grid=[[ax,ay,ax+swx,ay+swy],[linframe], $
	[xpos,ypos,xpos+xsize,ypos+swy],[a0,wedgemax,a0,wedgemax]]

return
end
