; ***********************************************************
; Actual convolution function
function myconvolve_weighted, map, weight, width, angle, aspect=aspect, init=init, outwgt=outwgt

if n_elements(aspect) ne 1 then aspect=sqrt(2.0)

common wvfields, xfeld, yfeld, fourierfeld, wfield, absfeld, dabs, fnx, fnxh, fny, fnyh, nmap


; FT map only if not done before
if keyword_set(init) then begin
  ; fix map if NANs contained
  invalid=where(~ finite(map),count)
  if count gt 0 then begin
    tmap=map
    tmap[invalid]=0.0
    fwdfourier, tmap, fourierfeld
  endif else fwdfourier, map, fourierfeld
  fwdfourier, weight, wfield
  ; field size
  ah=size(map)
  nx=ah[1]
  ny=ah[2]
  fnx=float(nx)
  fnxh=nx/2.0
  fny=float(ny)
  fnyh=ny/2.0
  yratio=fnxh/fnyh
  nmap=fnx*fny
  xfeld=fltarr(nx,ny)
  for i=0,nx-1 do xfeld[i,*]=replicate(i-fnxh,ny)
  yfeld=fltarr(nx,ny)
  for j=0,ny-1 do yfeld[*,j]=replicate(j-fnyh,nx)*yratio
  radiusfeld=sqrt(xfeld^2+yfeld^2)
  b=width*aspect
  bp=b*2.0*!pi/fnx
  absfeld=exp(-(radiusfeld*bp)^2/4.0)*nmap ;*!pi*b^2 
  bckfourier, wfield*absfeld, outwgt ; real
  outwgt=real_part(outwgt) > 0.0
  bckfourier, fourierfeld*absfeld, dabs ; real
  dabs=dabs/(outwgt >1e-5)
endif

; Now angle dependent part
pih=!pi/2.0
rotfeld=xfeld*cos(angle+pih)+yfeld*sin(angle+pih)

; define filter
a=width/(2.*!pi)
b=width*aspect
bp=b*2.0*!pi/fnx
ap=a*2.0*!pi/fnx
norm=exp(-b^2/(4.*a^2))
realfeld=absfeld*(cosh(rotfeld*bp^2/(2.0*ap))*exp(-bp^2/(4.*ap^2))-norm)
imagfeld=absfeld*sinh(rotfeld*bp^2/(2.0*ap))*exp(-bp^2/(4.*ap^2))
tfeld=realfeld+imagfeld
; Avoid NaNs from underruns
invalid=where(~ finite(tfeld),count)
if (count gt 0) and (count lt nmap) then begin
  tfeld[invalid]=0.0
  absfeld[invalid]=0.0
endif

; Convert back
bckfourier, wfield*tfeld, wconv ; complex
bckfourier, fourierfeld*tfeld, dconv ; complex
output1=dconv-dabs*wconv

; new version - should be correct in absolute scales
return, (abs(output1)^2)*b*aspect*!pi^2
end

; ***********************************************************
; Taper the edges if needed
pro taper, feld, wfeld, width=width, aspect=aspect, confidence=conf

if n_elements(aspect) ne 1 then aspect=sqrt(2.0)
if n_elements(conf) ne 1 then conf=0.9

; original field size
ah=size(feld)
if ah[0] gt 3 then begin
  print, "Field dimensions too high. Aborting"
  return
endif
if n_elements(width) ne 1 then width=min(ah[1:2])*0.02

; define weight field
if n_elements(wfeld) ne ah[4] then begin
  ; simple weights
  valid=where(finite(feld))
  wfeld=make_array(ah[1],ah[2],value=0.0)
  wfeld[valid]=1.0
endif

; Pad boundaries with zeros
extend=3.0
blankwidth=fix(width*extend+1)
weight=zeropad(wfeld, blankwidth)

; new field size
ah=size(weight)
nx=ah[1]
ny=ah[2]
fnx=float(nx)
fnxh=nx/2.0
fny=float(ny)
fnyh=ny/2.0
yratio=fnxh/fnyh
nmap=fnx*fny

; Initialize fields
; wasted effort, but duplication only once
xfeld=fltarr(nx,ny)
for i=0,nx-1 do xfeld[i,*]=replicate(i-fnxh,ny)
yfeld=fltarr(nx,ny)
for j=0,ny-1 do yfeld[*,j]=replicate(j-fnyh,nx)*yratio
radiusfeld=sqrt(xfeld^2+yfeld^2)

; Convolve, larger aspect needs wider tapering
b=width*aspect
bp=b*2.0*!pi/fnx
absfeld=exp(-(radiusfeld*bp)^2/4.0)*nmap
fwdfourier, weight,wdata
bckfourier, wdata*absfeld, absmag ; real
weight=((real_part(absmag)-conf) > 0.0)/(1.0-conf)
; Double convolve
fwdfourier, weight,wdata
bckfourier, wdata*absfeld, absmag ; real
iconf=(1.0-conf)/2.0
weight=((real_part(absmag)-iconf) > 0.0)/(1.0-iconf)

; cut back
wfeld=unpad(weight,blankwidth)
; taper field
feld=feld*wfeld

return
end

; ***********************************************************
; Caller - assumes weights are defined before
function waveletweighted, feld, wfeld, scale, aspect=aspect

if n_elements(aspect) ne 1 then aspect=sqrt(2.0)
; fixed parameter for extension width
extend=3.0

; angular grid
nang=180
angarr=findgen(nang)/nang*!pi

; Pad boundaries with zeros
blankwidth=fix(scale*extend+1)
dat=zeropad(feld, blankwidth)
weight=zeropad(wfeld, blankwidth)

; First call initilizing FT of main field
miarr=myconvolve_weighted(dat,weight,scale,0.0,aspect=aspect,/init,outwgt=outwgt)
marealarr=miarr
maimagarr=miarr*0.0

; all other angles
for i=1,nang-1 do begin &$
  rpart=cos(2*angarr[i]) &$
  ipart=sin(2*angarr[i]) &$
  output=myconvolve_weighted(dat,weight,scale,angarr[i],aspect=aspect) &$
  miarr=miarr+output &$
  marealarr=marealarr+output*rpart &$
  maimagarr=maimagarr+output*ipart &$
endfor

; Unpad if padded before
miarr=unpad(miarr,blankwidth)/nang
marealarr=unpad(marealarr,blankwidth)/nang
maimagarr=unpad(maimagarr,blankwidth)/nang
outw=unpad(outwgt, blankwidth)
absarr=sqrt(marealarr^2+maimagarr^2)
phiarr=atan(maimagarr,marealarr)/2.

return, [[[miarr]],[[absarr]],[[phiarr]],[[outw]]]
end

; ***********************************************************
; Caller - general case
; ***********************************************************
function waveletanalysis, map, scales, aspect=aspect, weights=weights, $
	edgetaper=edge, taperedmap=tmap, onephase=onephase, progress=progress

; edgetaper defines the edge treatment:
; edgetaper < 0 assumes that the map is periodic
; edgetaper = 0 assumes a non-periodic map that does not need edge tapering
; edgetaper > 0 (< 1) tapers the outer 2% of the map, 
; 		the tapered map is returned in taperedmap
; map is the input field to be analysed, invalid points should be NaN
; weights give the significance of each point in the map
;	they are only used in the non-period case: edgetaper >= 0
; scales is the array of wavelength scales for which the coefficients are
;	to be computed
; aspect gives the shape of the wavelet (b parameter)
; onephase can be "cos" or "sin" specifying that only the real or imaginary
;	phase is computed, if not given both are combined in complex plane 
; progress = 1 prints some status information while computing
; 
; return: 4 dimensional field:
; 	3-D array for each scale, consisting of 3-4 layers each 
;	1. layer: isotropic wavelet coefficients
;	2. layer: anisotropic wavelet coefficients
;	3. layer: argument of the anisotropic wavelet coefficients = angle
; 	4. layer: significance map per point from the convolution of the weights
;		(4. layer only for non-periodic maps)

; Map size
ah=size(map)
if ah[0] ne 2 then begin
  print, "Analysis can only deal with 2-D maps. Aborting"
  return, 0
endif
nx=ah[1]
ny=ah[2]

if (edge < 0) then begin
; Periodic version
  resfeld=fltarr(nx,ny,3,n_elements(scales))
  for i=0,n_elements(scales)-1 do begin 
    resfeld[*,*,*,i]=scaleconvolve(map,scales[i],aspect=aspect,onephase=onephase) 
    if progress then print, 'Scale: ', scales[i], ', mean coefficient ', $
		mean(resfeld[*,*,0,i],/nan) 
  endfor
endif else begin
if edge > 0 then begin
; Tapered-Version
  tmap=map
  taper, tmap, weights, confidence=edge
  resfeld=fltarr(nx,ny,4,n_elements(scales))
  for i=0,n_elements(scales)-1 do begin 
    resfeld[*,*,*,i]=waveletweighted(tmap,weights,scales[i],aspect=aspect) 
    if progress then print, 'Scale: ', scales[i], ', mean coefficient ', $
	mean(resfeld[*,*,0,i]*resfeld[*,*,3,i],/nan) 
  endfor
endif else begin
; simple weights
  if n_elements(weights) ne ah[4] then begin
    valid=where(finite(map))
    weights=make_array(nx,ny,value=0.0)
    weights[valid]=1.0
  endif
  resfeld=fltarr(nx,ny,4,n_elements(scales))
  for i=0,n_elements(scales)-1 do begin 
    resfeld[*,*,*,i]=waveletweighted(map,weights,scales[i],aspect=aspect) 
    if progress then print, 'Scale: ', scales[i], ', mean coefficient ', $
	mean(resfeld[*,*,0,i]*resfeld[*,*,3,i],/nan) 
  endfor
endelse
endelse

return, resfeld
end

